import bs
import bsUtils
import random
import bsBomb

_gFactory = None

defaultPowerupInterval = 8000


class PowerupMessage(object):
    def __init__(self,type,source=None):
        self.type = type
        self.source = source

class AcceptMessage(object):
    pass

class TouchedMessage(object):
    pass

class Factory(object):
    """ wraps up media an other resources used by spazzes """

    def __init__(self):

        #self.bombFactory = bsBomb.Factory()

        self._lastPowerupType = None

        self.model = bs.getModel("powerup")
        self.modelSimple = bs.getModel("powerupSimple")

        self.texBomb = bs.getTexture("powerupBomb")
        self.texSpeed = bs.getTexture("powerupSpeed")
        self.texPunch = bs.getTexture("powerupPunch")
        self.texIceBombs = bs.getTexture("powerupIceBombs")
        self.texStickyBombs = bs.getTexture("powerupStickyBombs")
        self.texShield = bs.getTexture("powerupShield")
        self.texImpactBombs = bs.getTexture("powerupImpactBombs")
        self.texHealth = bs.getTexture("powerupHealth")
        self.texLandMines = bs.getTexture("powerupLandMines")
        self.texCurse = bs.getTexture("powerupCurse")

        self.healthPowerupSound = bs.getSound("healthPowerup")
        self.powerupSound = bs.getSound("powerup01")
        self.powerdownSound = bs.getSound("powerdown01")
        self.dropSound = bs.getSound("boxDrop")

        # material for powerups
        self._material = bs._newMaterial()

        # material for anyone wanting to accept powerups
        self.powerupAcceptMaterial = bs._newMaterial()

        # pass a powerup-touched message to applicable stuff
        bs._materialAddAction(self._material,
                             conditions=(("dstIsMaterial",self.powerupAcceptMaterial)),
                             actions=(("partMod","collide",True),
                                      ("partMod","physical",False),
                                      ("message","self","atConnect",TouchedMessage())))
        # don't collide with newly spawned objects that don't accept us
        # bs._materialAddAction(self._material,
        #                      conditions=(("dstNotMaterial",self.powerupAcceptMaterial)),
        #                      actions=(("partMod","collide",True),
        #                               ("partMod","physical",False),
        #                               ("message","self","atConnect",TouchedMessage())))
        

        # we dont wanna be picked up
        bs._materialAddAction(self._material,
                             conditions=("dstIsMaterial",bs.pickupMaterial),
                                         actions=( ("partMod","collide",False) ) )

        bs._materialAddAction(self._material,
                         conditions=("dstIsMaterial",bs.footingMaterial),
                         actions=(("impactSound",self.dropSound,0.5,0.1)))

        self.powerupDist = []
        for p,freq in getDefaultPowerupDistribution():
            for i in range(int(freq)):
                self.powerupDist.append(p)

    def getRandomPowerupType(self,forceType=None,excludeTypes=[]):
        # if the last one was a curse, make this one a health to provide some hope
        if forceType: t = forceType
        else:
            if self._lastPowerupType == 'curse': t = 'health'
            else:
                while True:
                    t = self.powerupDist[random.randint(0,len(self.powerupDist)-1)]
                    if t not in excludeTypes:
                        break
        self._lastPowerupType = t
        return t

def getFactory():
    """ return the global factory object, initializing it if necessary """
    global _gFactory
    if _gFactory == None:
        _gFactory = Factory()
    return _gFactory


def getDefaultPowerupDistribution():
    return (("bomb",3),
            ("iceBombs",3),
            ("punch",3),
            ("impactBombs",3),
            ("landMines",2),
            ("stickyBombs",3),
            ("shield",2),
            ("health",1),
            ("curse",1))

class Powerup(bs.Actor):

    def __init__(self,activity,retain,translate=(0,1,0),powerupType="bomb",expire=True):
        bs.Actor.__init__(self,activity,retain)

        factory = getFactory()
        self.type = powerupType;
        self.powersGiven = False

        if powerupType == "bomb": tex = factory.texBomb
        elif powerupType == "speed": tex = factory.texSpeed
        elif powerupType == "punch": tex = factory.texPunch
        elif powerupType == "iceBombs": tex = factory.texIceBombs
        elif powerupType == "impactBombs": tex = factory.texImpactBombs
        elif powerupType == "landMines": tex = factory.texLandMines
        elif powerupType == "stickyBombs": tex = factory.texStickyBombs
        elif powerupType == "shield": tex = factory.texShield
        elif powerupType == "health": tex = factory.texHealth
        elif powerupType == "curse": tex = factory.texCurse
            
        self.node = bs.newNode("prop",
                               name='Powerup('+powerupType+')',
                               owner=activity,
                               delegate=self,
                               attrs={'body':"box",
                                      'translate':translate,
                                      'model':factory.model,
                                      'lightModel':factory.modelSimple,
                                      'shadowSize':0.5,
                                      'colorTexture':tex,
                                      'reflection':'powerup',
                                      'reflectionScale':[1.0],
                                      'materials':(factory._material,bs.objectMaterial)})

        # animate in..
        curve = bsUtils.animate(self.node,"modelScale",{0:0,140:1.6,200:1})
        bs.gameTimer(200,curve.delete)

        if expire:
            bs.gameTimer(defaultPowerupInterval-2500,bs.WeakCall(self.__startFlashing))
            bs.gameTimer(defaultPowerupInterval-1000,bs.WeakCall(self.handleMessage,bs.DieMessage()))


    def __startFlashing(self):
        if self.node.exists(): self.node.flashing = True

    def handleMessage(self,m):

        if isinstance(m,AcceptMessage):
            factory = getFactory()
            if self.type == 'health':
                bs.playSound(factory.healthPowerupSound,3,position=(self.node,"translate"))
            bs.playSound(factory.powerupSound,3,position=(self.node,"translate"))
            self.powersGiven = True
            self.handleMessage(bs.DieMessage())

        elif isinstance(m,TouchedMessage):
            if not self.powersGiven:
                node = bs.getCollideValues("opposingNode")
                if node.exists(): node.message(PowerupMessage(self.type,self.node))
                #bs.nodeMessage(bs.getCollideValues("opposingNode"),PowerupMessage(self.type,self.node))

        elif isinstance(m,bs.DieMessage):
            if self.node.exists():
                if (m.immediate):
                    self.node.delete()
                else:
                    curve = bsUtils.animate(self.node,"modelScale",{0:1,100:0})
                    bs.gameTimer(100,self.node.delete)

        elif isinstance(m,bs.OutOfBoundsMessage):
            self.handleMessage(bs.DieMessage())

        elif isinstance(m,bs.HitMessage):
            # dont die on punches (thats annoying)
            if m.hitType != 'punch':
                self.handleMessage(bs.DieMessage())
        else:
            bs.Actor.handleMessage(self,m)
