import bsCoopGame
import bsUtils
import bs
import bsSpaz
import bsBomb
import random
import bsAnnounce
import math
import bsUI
import bsMap
import bsScoreBoard
import weakref
import bsPowerup
import bsAchievement
import bsScoreSet

class RunaroundGame(bsCoopGame.CoopGameActivity):

    tips = ['Jump just as you\'re throwing to get bombs up to the highest levels.',
            'No, you can\'t get up on the ledge. You have to throw bombs.',
            'Whip back and forth to get more distance on your throws..']

    @classmethod
    def getName(c):
        return 'Runaround'

    @classmethod
    def getDescription(self,sessionType):
        return "Prevent enemies from reaching the exit."

    def __init__(self,session,settings={}):
        bsCoopGame.CoopGameActivity.__init__(self,session,'Tower D',settings)

        try: self._preset = self.settings['preset']
        except Exception: self._preset = 'pro'

        self._newWaveSound = bs.getSound('scoreHit01')
        self._winSound = bs.getSound("score")
        self._cashRegisterSound = bs.getSound('cashRegister')
        self._warnBeepsSound = bs.getSound("warnBeeps")
        self._badGuyScoreSound = bs.getSound("shieldDown")

        self._heartTex = bs.getTexture('heart')
        self._heartModelOpaque = bs.getModel('heartOpaque')
        self._heartModelTransparent = bs.getModel('heartTransparent')

        self._aPlayerHasBeenKilled = False

        self._spawnCenter = self._levelType.defs.points['spawn1'][0:3]
        self._tntSpawnPosition = self._levelType.defs.points['tntLoc'][0:3]
        self._powerupCenter = self._levelType.defs.boxes['powerupRegion'][0:3]
        self._powerupSpread = (self._levelType.defs.boxes['powerupRegion'][6]*0.5,self._levelType.defs.boxes['powerupRegion'][8]*0.5)

        self._scoreRegionMaterial = bs._newMaterial()
        bs._materialAddAction(self._scoreRegionMaterial,
                             conditions=("dstIsMaterial",bs.playerMaterial),
                             actions=(("partMod","collide",True),
                                      ("partMod","physical",False),
                                      ("call","atConnect",bs.WeakCall(self._handleReachedEnd))))

        self._lastWaveEndTime = bs.getGameTime()
        self._playerHasPickedUpPowerup = False

    def onTransitionIn(self):

        bsCoopGame.CoopGameActivity.onTransitionIn(self,music='Marching')

        self._level = self._levelType(self,bs.Actor.UNRETAINED)
        self._scoreBoard = bsScoreBoard.ScoreBoard(self,label=bs.getResource('scoreText'))
        self._gameOver = False
        self._wave = 0
        self._canEndWave = True
        self._score = 0
        self._timeBonus = 0

        self._scoreRegion = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                         "region",{"translate":self._level.defs.boxes["scoreRegion"][0:3],
                                                   "scale":self._level.defs.boxes["scoreRegion"][6:9],
                                                   "type": "box",
                                                   "materials":(self._scoreRegionMaterial,)})
        
    def onBegin(self):

        bsCoopGame.CoopGameActivity.onBegin(self,nameSound=bs.getSound('announceRunaround'))
        self._dingSound = bs.getSound('dingSmall')
        self._dingSoundHigh = bs.getSound('dingSmallHigh')

        playerCount = len(self.players)

        if self._preset == 'pro':
            self._excludePowerups = ['curse']
            self._haveTnt = True
            self._waves = [
                {'entries':[
                        {'type':bsSpaz.BomberBot,'path':3},
                        {'type':bsSpaz.BomberBot,'path':2},
                        {'type':bsSpaz.BomberBot,'path':2},
                        {'type':bsSpaz.BomberBot,'path':2} if playerCount > 1 else None,
                        {'type':bsSpaz.BomberBot,'path':1},
                        {'type':bsSpaz.BomberBot,'path':1} if playerCount > 2 else None,
                        {'type':bsSpaz.BomberBot,'path':1} if playerCount > 3 else None,
                        ]},
                {'entries':[
                        {'type':bsSpaz.BomberBot,'path':1},
                        {'type':bsSpaz.BomberBot,'path':2},
                        {'type':bsSpaz.BomberBot,'path':2},
                        {'type':bsSpaz.BomberBot,'path':2},
                        {'type':bsSpaz.BomberBot,'path':2} if playerCount > 3 else None,
                        {'type':bsSpaz.ToughGuyBot,'path':3},
                        {'type':bsSpaz.ToughGuyBot,'path':3},
                        {'type':bsSpaz.ToughGuyBot,'path':3},
                        {'type':bsSpaz.ToughGuyBot,'path':3} if playerCount > 1 else None,
                        {'type':bsSpaz.ToughGuyBot,'path':3} if playerCount > 2 else None,
                        ]},
                {'entries':[
                        {'type':bsSpaz.NinjaBot,'path':2},
                        {'type':bsSpaz.NinjaBot,'path':2} if playerCount > 2 else None,
                        {'type':bsSpaz.ChickBot,'path':2},
                        {'type':bsSpaz.ChickBot,'path':2} if playerCount > 1 else None,
                        {'type':'spacing','duration':3000},
                        {'type':bsSpaz.BomberBot,'path':2},
                        {'type':bsSpaz.BomberBot,'path':2},
                        {'type':bsSpaz.BomberBot,'path':2},
                        {'type':bsSpaz.BomberBot,'path':3},
                        {'type':bsSpaz.BomberBot,'path':3},
                        {'type':bsSpaz.BomberBot,'path':3},
                        {'type':bsSpaz.BomberBot,'path':3} if playerCount > 3 else None,
                        ]},
                {'entries':[
                        {'type':bsSpaz.ChickBot,'path':1},
                        {'type':'spacing','duration':1000},
                        {'type':bsSpaz.ChickBot,'path':2},
                        {'type':'spacing','duration':1000},
                        {'type':bsSpaz.ChickBot,'path':3},
                        {'type':'spacing','duration':1000},
                        {'type':bsSpaz.ChickBot,'path':1},
                        {'type':'spacing','duration':1000},
                        {'type':bsSpaz.ChickBot,'path':2},
                        {'type':'spacing','duration':1000},
                        {'type':bsSpaz.ChickBot,'path':3},
                        {'type':'spacing','duration':1000},
                        {'type':bsSpaz.ChickBot,'path':1} if playerCount > 1 else None,
                        {'type':'spacing','duration':1000},
                        {'type':bsSpaz.ChickBot,'path':2} if playerCount > 2 else None,
                        {'type':'spacing','duration':1000},
                        {'type':bsSpaz.ChickBot,'path':3} if playerCount > 3 else None,
                        {'type':'spacing','duration':1000},
                        ]},
                {'entries':[
                        {'type':bsSpaz.NinjaBotProShielded,'path':1},
                        {'type':bsSpaz.ToughGuyBot,'path':2},
                        {'type':bsSpaz.ToughGuyBot,'path':2},
                        {'type':bsSpaz.ToughGuyBot,'path':2},
                        {'type':bsSpaz.ToughGuyBot,'path':3},
                        {'type':bsSpaz.ToughGuyBot,'path':3},
                        {'type':bsSpaz.ToughGuyBot,'path':3},
                        {'type':bsSpaz.ToughGuyBot,'path':3} if playerCount > 1 else None,
                        {'type':bsSpaz.ToughGuyBot,'path':3} if playerCount > 2 else None,
                        {'type':bsSpaz.ToughGuyBot,'path':3} if playerCount > 3 else None,
                        ]},
                {'entries':[
                        {'type':bsSpaz.BomberBotProShielded,'path':3},
                        {'type':'spacing','duration':1500},
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':'spacing','duration':1500},
                        {'type':bsSpaz.BomberBotProShielded,'path':1},
                        {'type':'spacing','duration':1000},
                        {'type':bsSpaz.BomberBotProShielded,'path':3},
                        {'type':'spacing','duration':1500},
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':'spacing','duration':1500},
                        {'type':bsSpaz.BomberBotProShielded,'path':1},
                        {'type':'spacing','duration':1500},
                        {'type':bsSpaz.BomberBotProShielded,'path':3} if playerCount > 1 else None,
                        {'type':'spacing','duration':1500},
                        {'type':bsSpaz.BomberBotProShielded,'path':2} if playerCount > 2 else None,
                        {'type':'spacing','duration':1500},
                        {'type':bsSpaz.BomberBotProShielded,'path':1} if playerCount > 3 else None,
                        ]},
                ]
        elif self._preset == 'uber':
            self._excludePowerups = []
            self._haveTnt = True
            self._waves = [
                {'entries':[
                        {'type':bsSpaz.ChickBot,'path':1},
                        {'type':bsSpaz.ChickBot,'path':2},
                        {'type':bsSpaz.ChickBot,'path':2},
                        {'type':bsSpaz.ChickBot,'path':3},
                        {'type':bsSpaz.ToughGuyBotPro,'point':'BottomLeft'},
                        {'type':bsSpaz.ToughGuyBotPro,'point':'BottomRight'} if playerCount > 2 else None,
                        ]},
                {'entries':[
                        {'type':bsSpaz.NinjaBot,'path':2},
                        {'type':bsSpaz.NinjaBot,'path':3},
                        {'type':bsSpaz.NinjaBot,'path':1},
                        {'type':bsSpaz.NinjaBot,'path':2},
                        {'type':bsSpaz.NinjaBot,'path':3},
                        {'type':bsSpaz.NinjaBot,'path':1} if playerCount > 2 else None,
                        ]},
                {'entries':[
                        {'type':bsSpaz.BomberBotProShielded,'path':1},
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':bsSpaz.BomberBotProShielded,'path':3},
                        {'type':bsSpaz.BomberBotProShielded,'path':3},
                        {'type':bsSpaz.NinjaBot,'point':'BottomRight'},
                        {'type':bsSpaz.NinjaBot,'point':'BottomLeft'} if playerCount > 2 else None,
                        ]},
                {'entries':[
                        {'type':bsSpaz.ChickBotPro,'path':1},
                        {'type':bsSpaz.ChickBotPro,'path':1},
                        {'type':bsSpaz.ChickBotPro,'path':1},
                        {'type':bsSpaz.ChickBotPro,'path':1},
                        {'type':bsSpaz.ChickBotPro,'path':1},
                        {'type':bsSpaz.ChickBotPro,'path':1},
                        {'type':bsSpaz.ChickBotPro,'path':1} if playerCount > 1 else None,
                        {'type':bsSpaz.ChickBotPro,'path':1} if playerCount > 3 else None,
                        ]},
                {'entries':[
                        {'type':bsSpaz.ChickBotProShielded,'point':'BottomLeft'},
                        {'type':bsSpaz.ChickBotProShielded,'point':'BottomRight'},
                        {'type':bsSpaz.ChickBotProShielded,'point':'BottomRight'} if playerCount > 2 else None,
                        {'type':bsSpaz.BomberBot,'path':3},
                        {'type':bsSpaz.BomberBot,'path':3},
                        {'type':'spacing','duration':5000},
                        {'type':bsSpaz.ToughGuyBot,'path':2},
                        {'type':bsSpaz.ToughGuyBot,'path':2},
                        {'type':'spacing','duration':5000},
                        {'type':bsSpaz.ChickBot,'path':1},
                        {'type':bsSpaz.ChickBot,'path':1},
                        ]},
                {'entries':[
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':bsSpaz.MelBot,'point':'BottomRight'},
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':bsSpaz.MelBot,'point':'BottomRight'} if playerCount > 2 else None,
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':bsSpaz.PirateBot,'point':'BottomLeft'},
                        {'type':bsSpaz.BomberBotProShielded,'path':2},
                        {'type':bsSpaz.BomberBotProShielded,'path':2} if playerCount > 1 else None,
                        {'type':'spacing','duration':5000},
                        {'type':bsSpaz.MelBot,'point':'BottomLeft'},
                        {'type':'spacing','duration':2000},
                        {'type':bsSpaz.PirateBot,'point':'BottomRight'},
                        ]},
                ]
        elif self._preset == 'endless':
            self.createControlsHelpUI(durationScale=0.3)
            self._excludePowerups = []
            self._haveTnt = True
            

        # spit out a few powerups and start dropping more shortly
        self._dropPowerups(standardPoints=True)
        bs.gameTimer(4000,bs.WeakCall(self._startPowerupDrops))
        self.createLifeWarningTimer()
        self._updateScores()
        self._bots = bsSpaz.BotSet(self.players)

        # our TNT spawner (if applicable)
        if self._haveTnt: self._tntSpawner = bsBomb.TNTSpawner(activity=self,position=self._tntSpawnPosition)

        lOffs = -20 if bs.getInterfaceType() == 'phone' else 0

        self._livesBG = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                     'image',{'texture':self._heartTex,
                                              'modelOpaque':self._heartModelOpaque,
                                              'modelTransparent':self._heartModelTransparent,
                                              'attach':'topRight',
                                              'scale':(90,90,90),
                                              'translate':(-100+lOffs,-50),
                                              'color':(1,0.2,0.2)},name="LivesBG")

        self._startLives = 10
        self._lives = self._startLives
        self._livesText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                       'text',{'vAttach':'top',
                                               'hAttach':'right',
                                               'hAlign':'center',
                                               'color':(0.8,0.8,0.8,1.0),
                                               'translate':(-103+lOffs,-63),
                                               'scale':[1.3],
                                               'text':str(self._lives)},
                                       name='Lives Text')

        bs.gameTimer(2000,bs.WeakCall(self._startUpdatingWaves))

    def _handleReachedEnd(self):
        n = bs.getCollideValues("opposingNode")
        spaz = n.getDelegate()

        if spaz.dead: return # ignore bodies flying in..

        self._flawless = False
        p = spaz.node.translate
        bs.playSound(self._badGuyScoreSound,position=p)
        light = bs.newNode("light",
                           owner=self,
                           attrs={"translate": p,
                                  "radius":0.5,
                                  "color": (1,0,0)})
        bsUtils.animate(light,"intensity",{0:0,100:1,500:0},loop=False)
        bs.gameTimer(1000,light.delete)
        spaz.handleMessage(bs.DieMessage(immediate=True,how='goal'))

        if self._lives > 0:
            self._lives -= 1
            if self._lives == 0:
                bs.gameTimer(1,bs.WeakCall(self.doEnd,'defeat'))
                bsUtils.playMusic(None)
                bs.playSound(bsSpaz.getFactory().singlePlayerDeathSound)

            self._livesText.node.text = str(self._lives)
            t = 0
            def _safeSetAttr(node,attr,value):
                if node.exists(): setattr(node,attr,value)
            for i in range(4):
                bs.gameTimer(t,bs.WeakCall(_safeSetAttr,self._livesText.node,'color',(1,0,0,1.0)))
                bs.gameTimer(t,bs.WeakCall(_safeSetAttr,self._livesBG.node,'opacity',0.5))
                t += 125
                bs.gameTimer(t,bs.WeakCall(_safeSetAttr,self._livesText.node,'color',(1.0,1.0,0.0,1.0)))
                bs.gameTimer(t,bs.WeakCall(_safeSetAttr,self._livesBG.node,'opacity',1.0))
                t += 125
            bs.gameTimer(t,bs.WeakCall(_safeSetAttr,self._livesText.node,'color',(0.8,0.8,0.8,1.0)))
            

    def spawnPlayer(self,player):
        pos = (self._spawnCenter[0]+random.uniform(-1.5,1.5),self._spawnCenter[1],self._spawnCenter[2]+random.uniform(-1.5,1.5))
        s = self.spawnSpazForPlayer(player,position=pos)

        # add the material that causes us to hit the player-wall
        for attr in ['materials','rollerMaterials']:
            materials = getattr(s.node,attr)
            setattr(s.node,attr,materials + (self._level.collideWithWallMaterial,))

        s.pickUpPowerupCallback = bs.WeakCall(self._handlePlayerPickedUpPowerup)

    def _handlePlayerPickedUpPowerup(self,player):
        self._playerHasPickedUpPowerup = True

    def _dropPowerup(self,index,powerupType=None):
        if powerupType == None: powerupType = bsPowerup.getFactory().getRandomPowerupType(excludeTypes=self._excludePowerups)
        bsPowerup.Powerup(self,bs.Actor.RETAINED,translate=self._level.powerupSpawnPoints[index],powerupType=powerupType)

    def _startPowerupDrops(self):
        self._powerupDropTimer = bs.Timer(3000,bs.WeakCall(self._dropPowerups),repeat=True)

    def _dropPowerups(self,standardPoints=False,forceFirst=None):
        """ Generic powerup drop """

        # if its been a minute since our last wave finished emerging, stop giving out land-mine powerups.
        # (prevents players from waiting around for them on purpose and filling the map up)
        if bs.getGameTime() - self._lastWaveEndTime > 60000: extraExcludes = ['landMines']
        else: extraExcludes = []

        if standardPoints:
            pts = self._level.powerupSpawnPoints
            for i,pt in enumerate(pts):
                bs.gameTimer(1000+i*500,bs.WeakCall(self._dropPowerup,i,forceFirst if i == 0 else None))
        else:
            pt = (self._powerupCenter[0]+random.uniform(-1.0*self._powerupSpread[0],1.0*self._powerupSpread[0]),
                  self._powerupCenter[1],self._powerupCenter[2]+random.uniform(-self._powerupSpread[1],self._powerupSpread[1]))

            # drop one random one somewhere..
            bsPowerup.Powerup(self,bs.Actor.RETAINED,translate=pt,powerupType=bsPowerup.getFactory().getRandomPowerupType(excludeTypes=self._excludePowerups+extraExcludes))


    def doEnd(self,outcome):

        if outcome == 'defeat':
            delay = 2000
            self.fadeToRed()
        else: delay = 0

        if self._wave >= 2:
            score = self._score
            failMessage = None
        else:
            score = None
            failMessage = 'Reach wave 2 to rank.'

        self.end(delay=delay,results={'outcome':outcome,'score':score,'failMessage':failMessage,'playerInfo':self.initialPlayerInfo})

    def onGotScoresToBeat(self,scores):
        self.showStandardScoresToBeatUI(scores)

    def _updateWaves(self):

        # if we have no living bots, go to the next wave
        if self._canEndWave and not self._bots.haveLivingBots() and not self._gameOver and self._lives > 0:

            self._canEndWave = False

            self._timeBonusTimer = None
            self._timeBonusText = None

            if self._preset == 'endless': won = False
            else: won = (self._wave == len(self._waves))

            # reward time bonus
            baseDelay = 4000 if won else 0
            if self._timeBonus > 0:
                bs.gameTimer(0,bs.WeakCall(bs.playSound,self._cashRegisterSound))
                bs.gameTimer(baseDelay,bs.WeakCall(self._awardTimeBonus,self._timeBonus))
                baseDelay += 1000

            # reward flawless bonus
            if self._wave > 0 and self._flawless:
                bs.gameTimer(baseDelay,bs.WeakCall(self._awardFlawlessBonus))
                baseDelay += 1000

            self._flawless = True # reset

            if won:

                # completion achievements
                if self._preset == 'pro':
                    self.awardAchievement('Pro Runaround Victory',sound=False)
                    if self._lives == self._startLives: self.awardAchievement('The Wall',sound=False)
                    if not self._playerHasPickedUpPowerup: self.awardAchievement('Precision Bombing',sound=False)
                        
                elif self._preset == 'uber':
                    self.awardAchievement('Uber Runaround Victory',sound=False)
                    if self._lives == self._startLives: self.awardAchievement('The Great Wall',sound=False)
                    if not self._aPlayerHasBeenKilled: self.awardAchievement('Stayin\' Alive',sound=False)

                # give remaining players some points and have them celebrate
                self._victoryText = bsAnnounce.ZoomText(self,bs.Actor.UNRETAINED,
                                                        bs.getResource('victoryText'),maxWidth=800,lifespan=4000,jitter=2.0,position=(0,200))
                self.celebrate(10000)

                bs.gameTimer(baseDelay,bs.WeakCall(self._awardLivesBonus))
                baseDelay += 1000
                
                bs.gameTimer(baseDelay,bs.WeakCall(self._awardCompletionBonus))
                baseDelay += 850

                bs.playSound(self._winSound)
                self.cameraFlash()
                bsUtils.playMusic('Victory')
                self._gameOver = True
                bs.gameTimer(baseDelay,bs.WeakCall(self.doEnd,'victory'))
                return

            self._wave += 1

            # short celebration after waves
            if self._wave > 1: self.celebrate(500)
                            
            bs.gameTimer(baseDelay,bs.WeakCall(self._startNextWave))

    def _awardCompletionBonus(self):
        bonus = 200
        bs.playSound(self._cashRegisterSound)
        bsAnnounce.AchievementText(self,bs.Actor.RETAINED,
                                   '+'+str(bonus)+' '+bs.getResource('completionBonusText'),
                                   color=(0.7,0.7,1.0,1),
                                   scale=1.6,
                                   position=(0,1.5,-1))
        self._score += bonus
        self._updateScores()

    def _awardLivesBonus(self,):
        bonus = self._lives * 30
        bs.playSound(self._cashRegisterSound)
        bsAnnounce.AchievementText(self,bs.Actor.RETAINED,
                                   '+'+str(bonus)+' '+bs.getResource('livesBonusText'),
                                   color=(0.7,1.0,0.3,1),
                                   scale=1.3,
                                   position=(0,1,-1))
        self._score += bonus
        self._updateScores()

    def _awardTimeBonus(self,bonus):
        bs.playSound(self._cashRegisterSound)
        bsAnnounce.AchievementText(self,bs.Actor.RETAINED,
                                   '+'+str(bonus)+' '+bs.getResource('timeBonusText'),
                                   color=(1,1,0.5,1),
                                   scale=1.0,
                                   position=(0,3,-1))
        self._score += self._timeBonus
        self._updateScores()

    def _awardFlawlessBonus(self):
        bs.playSound(self._cashRegisterSound)
        bsAnnounce.AchievementText(self,bs.Actor.RETAINED,
                                   '+'+str(self._flawlessBonus)+' '+bs.getResource('perfectWaveText'),
                                   color=(1,1,0.2,1),
                                   scale=1.2,
                                   position=(0,2,-1))
        self._score += self._flawlessBonus
        self._updateScores()

    def _startTimeBonusTimer(self):
        self._timeBonusTimer = bs.Timer(1000,bs.WeakCall(self._updateTimeBonus),repeat=True)
        

    def _startNextWave(self):

        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,
                            bs.getResource('waveText')+' '+str(self._wave),maxWidth=800,lifespan=1000,jitter=2.0,position=(0,200))
        bs.gameTimer(400,bs.WeakCall(bs.playSound,self._newWaveSound))
        t = 0
        baseDelay = 500
        botAngle = random.random()*360.0
        spawnTime = 100
        botTypes = []

        if self._preset == 'endless':

            level = self._wave
            targetPoints = (level+1) * 8.0
            groupCount = random.randint(1,3)
            entries = []

            spazTypes = []
            if level < 6: spazTypes += [[bsSpaz.BomberBot,5.0]]
            if level < 10: spazTypes += [[bsSpaz.ToughGuyBot,5.0]]
            if level < 15: spazTypes += [[bsSpaz.ChickBot,6.0]]
            if level > 5: spazTypes += [[bsSpaz.ChickBotPro,7.5]]*(1+(level-5)/7)
            if level > 2: spazTypes += [[bsSpaz.BomberBotProShielded,8.0]]*(1+(level-2)/6)
            if level > 6: spazTypes += [[bsSpaz.ChickBotProShielded,12.0]]*(1+(level-6)/5)
            if level > 1: spazTypes += [[bsSpaz.NinjaBot,10.0]]*(1+(level-1)/4)
            if level > 7: spazTypes += [[bsSpaz.NinjaBotProShielded,15.0]]*(1+(level-7)/3)

            # bot type, their effect on target points
            defenderTypes = [[bsSpaz.BomberBot,0.9],
                             [bsSpaz.ToughGuyBot,0.9],
                             [bsSpaz.ChickBot,0.85]]
            if level > 2: defenderTypes += [[bsSpaz.NinjaBot,0.75]]
            if level > 4: defenderTypes += [[bsSpaz.MelBot,0.7]]*(1+(level-5)/6)
            if level > 6: defenderTypes += [[bsSpaz.PirateBot,0.7]]*(1+(level-5)/5)
            if level > 8: defenderTypes += [[bsSpaz.ToughGuyBotProShielded,0.65]]*(1+(level-5)/4)
            if level > 10: defenderTypes += [[bsSpaz.ChickBotProShielded,0.6]]*(1+(level-6)/3)
                             
            for group in range(groupCount):
                thisTargetPoints = targetPoints/groupCount

                # adding spacing makes things slightly harder
                r = random.random()
                if r < 0.07:
                    spacing = 1500
                    thisTargetPoints *= 0.85
                elif r < 0.15:
                    spacing = 1000
                    thisTargetPoints *= 0.9
                else: spacing = 0

                path = random.randint(1,3)
                # dont allow hard paths on early levels
                if level < 3:
                    if path == 1: path = 3
                # easy path
                if path == 3: pass
                # harder path
                elif path == 2: thisTargetPoints *= 0.8
                # even harder path
                elif path == 1: thisTargetPoints *= 0.7
                # looping forward
                elif path == 4: thisTargetPoints *= 0.7
                # looping backward
                elif path == 5: thisTargetPoints *= 0.7
                # random
                elif path == 6: thisTargetPoints *= 0.7

                def _addDefender(defenderType,point):
                    #entries.append()
                    return thisTargetPoints*defenderType[1],{'type':defenderType[0],'point':point}

                # add defenders
                defenderType1 = defenderTypes[random.randrange(len(defenderTypes))]
                defenderType2 = defenderTypes[random.randrange(len(defenderTypes))]
                defender1 = defender2 = None
                if (group == 0) or (group == 1 and level > 3) or (group == 2 and level > 5):
                    if random.random() < min(0.75,(level-1)*0.11):
                        thisTargetPoints,defender1 = _addDefender(defenderType1,'BottomLeft')
                    if random.random() < min(0.75,(level-1)*0.04):
                        thisTargetPoints,defender2 = _addDefender(defenderType2,'BottomRight')


                spazType = spazTypes[random.randrange(len(spazTypes))]
                memberCount = max(1,int(round(thisTargetPoints/spazType[1])))
                for i,member in enumerate(range(memberCount)):
                    if path == 4: thisPath = i%3 # looping forward
                    elif path == 5: thisPath = 3-(i%3) # looping backward
                    elif path == 6: thisPath = random.randint(1,3) # random
                    else: thisPath = path
                    entries.append({'type':spazType[0],'path':thisPath})
                    if spacing != 0:
                        entries.append({'type':'spacing','duration':spacing})

                if defender1 is not None: entries.append(defender1)
                if defender2 is not None: entries.append(defender2)

                # some spacing between groups
                r = random.random()
                if r < 0.1: spacing = 5000
                elif r < 0.5: spacing = 1000
                else: spacing = 1
                entries.append({'type':'spacing','duration':spacing})

            wave =  {'entries':entries}

        else:
            wave = self._waves[self._wave-1]

        try: botAngle = wave['baseAngle']
        except Exception: botAngle = 0

        botTypes += wave['entries']

        self._timeBonusMult = 1.0
        thisFlawlessBonus = 0

        nonRunnerSpawnTime = 1000
        for info in botTypes:
            if info == None: continue

            botType = info['type']

            if botType != None:

                if botType == 'nonRunnerDelay':
                    nonRunnerSpawnTime += info['duration']
                    continue

                if botType == 'spacing':
                    t += info['duration']
                    continue
                else:
                    try: path = info['path']
                    except Exception: path = random.randint(1,3)
                    self._timeBonusMult += botType.pointsMult * 0.02
                    thisFlawlessBonus += botType.pointsMult * 5

            # if its got a position, use that
            try: point = info['point']
            except Exception: point = 'Start'

            # space our our slower bots
            delay = baseDelay
            delay /= self._getBotSpeed(botType)

            t += int(delay*0.5)
            bs.gameTimer(t,bs.WeakCall(self.addBotAtPoint,point,{'type':botType,'path':path},100 if point == 'Start' else nonRunnerSpawnTime))
            t += int(delay*0.5)

        # we can end the wave after all the spawning happens
        bs.gameTimer(t-int(delay*0.5)+nonRunnerSpawnTime+10,bs.WeakCall(self._setCanEndWave))

        # reset our time bonus
        # in this game we use a constant time bonus so it erodes away in roughly the same time
        # (since the time limit a wave can take is relatively constant)
        # ..we then post-multiply a modifier to adjust points
        self._timeBonus = 150
        self._flawlessBonus = thisFlawlessBonus

        self._timeBonusText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                           'text',{'vAttach':'top',
                                                   'hAttach':'center',
                                                   'hAlign':'center',
                                                   'color':(1,1,0.5,1),
                                                   'translate':(0,-60),
                                                   'scale':[0.6],
                                                   'text':bs.getResource('timeBonusText')+': '+str(int(self._timeBonus*self._timeBonusMult))},
                                           name="Time Bonus Left")
        bs.gameTimer(t,bs.WeakCall(self._startTimeBonusTimer))

        # keep track of when this wave finishes emerging - we wanna
        # stop dropping land-mines powerups at some point
        # (otherwise a crafty player could fill the whole map with them)
        self._lastWaveEndTime = bs.getGameTime()+t
        self._waveText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                      'text',{'vAttach':'top',
                                              'hAttach':'center',
                                              'hAlign':'center',
                                              'color':(0.6,0.6,0.6,1.0),
                                              'translate':(0,-40),
                                              'scale':[1.1],
                                              'text':bs.getResource('waveText')+' '+str(self._wave)+('' if self._preset == 'endless' else ('/'+str(len(self._waves))))},
                                      name='Wave Number')


    def addBotAtPoint(self,point,spazInfo,spawnTime=100):
        # dont add if the game has ended
        if self._gameOver: return
        pt = self._level.defs.points['botSpawn'+point]
        bsUtils.Spawner(activity=self,data=spazInfo,pt=pt,spawnTime=spawnTime)
        


    def _updateTimeBonus(self):
        self._timeBonus = int(self._timeBonus * 0.91)
        if self._timeBonus > 0 and self._timeBonusText != None:
            self._timeBonusText.node.text = bs.getResource('timeBonusText')+': '+str(int(self._timeBonus*self._timeBonusMult))
        else: self._timeBonusText = None

    def _startUpdatingWaves(self):
        self._waveUpdateTimer = bs.Timer(2000,bs.WeakCall(self._updateWaves),repeat=True)
        
    def _updateScores(self):

        if self._preset == 'endless':
            if self._score >= 500: self.awardAchievement('Runaround Master')
            if self._score >= 1000: self.awardAchievement('Runaround Wizard')
            if self._score >= 2000: self.awardAchievement('Runaround God')

        self._scoreBoard.setTeamValue(self.teams[0],self._score,maxScore=None)
        
    def _updateBot(self,bot):

        speed = bot.rWalkSpeed
        t = bot.node.translate

        boxes = self._level.defs.boxes

        # bots in row 1 attempt the high road..
        if bot.rWalkRow == 1:
            if bsUtils.isPointInBox(t,boxes['b4']):
                bot.node.message('moveUpDown',speed * 32000)
                bot.node.message('moveLeftRight',0)
                bot.node.message('runRelease')
                return True

        # row 1 and 2 bots attempt the middle road..
        if bot.rWalkRow in [1,2]:
            if bsUtils.isPointInBox(t,boxes['b1']):
                bot.node.message('moveUpDown',speed * 32000)
                bot.node.message('moveLeftRight',0)
                bot.node.message('runRelease')
                return True

        # *all* bots settle for the third row
        if bsUtils.isPointInBox(t,boxes['b7']):
            bot.node.message('moveUpDown',speed * 32000)
            bot.node.message('moveLeftRight',0)
            bot.node.message('runRelease')
            return True
        elif bsUtils.isPointInBox(t,boxes['b2']):
            bot.node.message('moveUpDown',-speed * 32000)
            bot.node.message('moveLeftRight',0)
            bot.node.message('runRelease')
            return True
        elif bsUtils.isPointInBox(t,boxes['b3']):
            bot.node.message('moveUpDown',-speed * 32000)
            bot.node.message('moveLeftRight',0)
            bot.node.message('runRelease')
            return True
        elif bsUtils.isPointInBox(t,boxes['b5']):
            bot.node.message('moveUpDown',-speed * 32000)
            bot.node.message('moveLeftRight',0)
            bot.node.message('runRelease')
            return True
        elif bsUtils.isPointInBox(t,boxes['b6']):
            bot.node.message('moveUpDown',speed * 32000)
            bot.node.message('moveLeftRight',0)
            bot.node.message('runRelease')
            return True
        elif (bsUtils.isPointInBox(t,boxes['b8']) and not bsUtils.isPointInBox(t,boxes['b9'])) or t == (0.0,0.0,0.0):
            # default to walking right if we're still in the walking area
            # this also catches bots that havnt been positioned yet (translate ==(0,0,0))
            bot.node.message('moveLeftRight',speed * 32000)
            bot.node.message('moveUpDown',0)
            bot.node.message('runRelease')
            return True

        # revert to normal bot behavior otherwise..
        return False

    def handleMessage(self,m):

        if isinstance(m,bsScoreSet.PlayerScoredMessage):
            self._score += m.score
            self._updateScores()

        # respawn dead players
        elif isinstance(m,bsSpaz.PlayerSpazDeathMessage):
            self._aPlayerHasBeenKilled = True
            player = m.spaz.getPlayer()
            if player is None: return

            self.scoreSet.playerLostSpaz(player)

            # respawn them shortly
            respawnTime = 2000+len(self.initialPlayerInfo)*1000

            player.gameData['respawnTimer'] = bs.Timer(respawnTime,bs.WeakCall(self.spawnPlayerRef,weakref.ref(player)))
            player.gameData['respawnIcon'] = bsSpaz.RespawnIcon(player,self,respawnTime)

        elif isinstance(m,bsSpaz.SpazBotDeathMessage):
            if m.how == 'goal': return
            pts,importance = m.badGuy.getDeathPoints(m.how)
            if m.killer != None:
                try: target = m.badGuy.node.translate
                except Exception: target = None
                try:
                    killerPlayer = self.players[m.killer]
                    self.scoreSet.playerScored(killerPlayer,pts,target=target,kill=True,screenMessage=False,importance=importance)
                    bs.playSound(self._dingSound if importance == 1 else self._dingSoundHigh,volume=0.6)
                except Exception: pass
            # normally we pull scores from the score-set, but if there's no player lets be explicit..
            else: self._score += pts
            self._updateScores()

        elif isinstance(m,bsUtils.Spawner.SpawnMessage):

            info = m.data
            spazType = info['type']
            path = info['path']
            spaz = spazType(self,bs.Actor.UNRETAINED,level=self._level)
            
            spaz.updateCallback = bs.WeakCall(self._updateBot)

            # override the time limit on pirates..
            # if spazType == bsSpaz.PirateBot:
            #     spaz.node.curseDeathTime = -1

            spaz.rWalkRow = path
            spaz.rWalkSpeed = self._getBotSpeed(spazType)

            # add our custom update callback
            bs.playSound(self._spawnSound,position=m.pt)
            spaz.node.message("flash")
            spaz.node.isAreaOfInterest = 0
            spaz.handleMessage(bs.StandMessage(m.pt,random.uniform(0,360)))

            light = bs.newNode("light",
                               owner=self,
                               attrs={"translate": m.pt,
                                      "radius":0.4,
                                      "color": (1,0,0)})
            bsUtils.animate(light,"intensity",{0:0,100:1,500:0},loop=False)
            bs.gameTimer(1000,light.delete)

            self._bots.addBot(spaz)

        else:
            self.__superHandleMessage(m)

    def __superHandleMessage(self,m):
        super(RunaroundGame,self).handleMessage(m)

    def _getBotSpeed(self,botType):
        if botType == bsSpaz.BomberBot: return 0.48
        elif botType == bsSpaz.BomberBotPro: return 0.48
        elif botType == bsSpaz.BomberBotProShielded: return 0.48
        elif botType == bsSpaz.ToughGuyBot: return 0.57
        elif botType == bsSpaz.ToughGuyBotPro: return 0.57
        elif botType == bsSpaz.ToughGuyBotProShielded: return 0.57
        elif botType == bsSpaz.ChickBot: return 0.73
        elif botType == bsSpaz.ChickBotPro: return 0.78
        elif botType == bsSpaz.ChickBotProShielded: return 0.78
        elif botType == bsSpaz.NinjaBot: return 1.0
        elif botType == bsSpaz.NinjaBotProShielded: return 1.0
        elif botType == bsSpaz.PirateBot: return 1.0
        elif botType == bsSpaz.MelBot: return 0.5
        else: raise Exception('Invalid bot type to _getBotSpeed(): '+str(botType))
        
    def _setCanEndWave(self):
        self._canEndWave = True
